/**
 *   @file  em_pdk_hal.c
 *
 *   @brief   
 *      This file contains the routines to abstract some of CPPI/QMSS drivers functions
 *      for the event machine.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2012, Texas Instruments, Inc.
 *  ============================================================================ 
 *  \par
 */
#include <stdio.h>
#include <string.h>

/* importing some of the Open EM defined symbols */
#include <ti/runtime/openem/event_machine.h>

#include <ti/csl/csl_srio.h>
#include <ti/csl/csl_srioAux.h>

#include <ti/drv/cppi/cppi_drv.h>
#include <ti/drv/cppi/cppi_desc.h>
#include <ti/drv/cppi/include/cppi_pvt.h>
#include <ti/drv/qmss/qmss_drv.h>
#include <ti/drv/srio/srio_drv.h>

#define XGE_FLOW_IDX_START 20

/*
 * Arrays to store the CPPI PDK handles
 */
Cppi_FlowHnd flowHandleArray[3][64];
Cppi_ChHnd rxChannelHandleArray[3][32];
Cppi_ChHnd txChannelHandleArray[3][32];

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_hw_queue_open (
    int queue_idx
)
{
    Qmss_QueueType lvQueueType;
    Qmss_QueueHnd lvQueueHandle;
    uint16_t lvRelativeQueueIdx = queue_idx & 0x1FFF;
    uint8_t lvIsAllocated;

    if ((QMSS_SRIO_QUEUE_BASE <= queue_idx) && (queue_idx < (QMSS_SRIO_QUEUE_BASE + QMSS_MAX_SRIO_QUEUE)))
    {
        lvQueueType = Qmss_QueueType_SRIO_QUEUE;
    }
    else if ((lvRelativeQueueIdx >= (QMSS_INFRASTRUCTURE_QUEUE_BASE)) &&
             (lvRelativeQueueIdx < (QMSS_INFRASTRUCTURE_QUEUE_BASE + QMSS_MAX_INFRASTRUCTURE_QUEUE)))
    {
        lvQueueType = Qmss_QueueType_INFRASTRUCTURE_QUEUE;
    }
    else
    {
        lvQueueType = Qmss_QueueType_GENERAL_PURPOSE_QUEUE;
    }

    lvQueueHandle = Qmss_queueOpen (lvQueueType, queue_idx, &lvIsAllocated);
    // The queue shall not be already used by another application.
    if ((lvQueueHandle < 0) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_hw_queue_close (
    int queue_idx
)
{
    Qmss_Result lvRes;
    // Empty the queue
    Qmss_queueEmpty ((Qmss_QueueHnd) queue_idx);
    // Close the queue
    lvRes = Qmss_queueClose ((Qmss_QueueHnd) queue_idx);
    if (lvRes != 0)
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

static inline int8_t
ti_em_retrieve_array_index (
    Cppi_CpDma dmaNum
)
{
    int8_t lvArrayIdx;

    switch (dmaNum)
    {
        case Cppi_CpDma_QMSS_CPDMA:
            lvArrayIdx = 0;
            break;
        case Cppi_CpDma_SRIO_CPDMA:
            lvArrayIdx = 1;
            break;
#ifdef TI_EM_C6634
        case Cppi_CpDma_QMSS_QM2_CPDMA:
            lvArrayIdx = 2;
            break;
#endif
        default:
            /*
             * Invalid CPPI
             */
            lvArrayIdx = -1;
            break;
    }
    return lvArrayIdx;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idx,
    int error_handling
)
{
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_dest_tag_lo_sel = 4;
    lvRxFlowCfg.rx_dest_tag_hi_sel = 5;
    lvRxFlowCfg.rx_src_tag_hi_sel = 4;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq1_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq1_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq2_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq2_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq3_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq3_qmgr = (free_queue_idx >> 12);
    /*
     * Other fields are left to be modified by user 
     */

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    /*
     * Store the handle for future usage
     */
    flowHandleArray[lvArrayIdx][flow_idx] = lvRxFlowHnd;

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_flow_close (
    int dma_idx,
    int flow_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    Cppi_Result lvRes;
    int lvFlowIdx = flow_idx;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxFlowHnd = flowHandleArray[lvArrayIdx][lvFlowIdx];
    // Workaround for XGE current testing
    if (lvRxFlowHnd == (Cppi_FlowHnd) 0)
    {
        lvFlowIdx += XGE_FLOW_IDX_START;
        lvRxFlowHnd = flowHandleArray[lvArrayIdx][lvFlowIdx];
    }
    lvRes = Cppi_closeRxFlow (lvRxFlowHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    flowHandleArray[lvArrayIdx][lvFlowIdx] = (Cppi_FlowHnd) 0;
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idx0,
    int free_queue_idx1,
    int error_handling
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = XGE_FLOW_IDX_START + flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_dest_tag_lo_sel = 4;
    lvRxFlowCfg.rx_dest_tag_hi_sel = 5;
    lvRxFlowCfg.rx_src_tag_hi_sel = 4;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idx0 & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idx0 >> 12);
    lvRxFlowCfg.rx_fdq1_qnum = (free_queue_idx1 & 0xFFF);
    lvRxFlowCfg.rx_fdq1_qmgr = (free_queue_idx1 >> 12);
    lvRxFlowCfg.rx_fdq2_qnum = (free_queue_idx1 & 0xFFF);
    lvRxFlowCfg.rx_fdq2_qmgr = (free_queue_idx1 >> 12);
    lvRxFlowCfg.rx_fdq3_qnum = (free_queue_idx1 & 0xFFF);
    lvRxFlowCfg.rx_fdq3_qmgr = (free_queue_idx1 >> 12);



    /*
     * Other fields are left to be modified by user 
     */

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;

    flowHandleArray[lvArrayIdx][lvRxFlowCfg.flowIdNum] = lvRxFlowHnd;

    return (em_status_t) EM_OK;

}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_flow_close (
    int dma_idx,
    int flow_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    Cppi_Result lvRes;
    int lvFlowIdx = flow_idx + XGE_FLOW_IDX_START;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxFlowHnd = flowHandleArray[lvArrayIdx][lvFlowIdx];
    lvRes = Cppi_closeRxFlow (lvRxFlowHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    flowHandleArray[lvArrayIdx][lvFlowIdx] = (Cppi_FlowHnd) 0;
    return (em_status_t) EM_OK;
}


/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rio_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idxSizeA,
    int sizeA,
    int free_queue_idxSizeB,
    int sizeB,
    int free_queue_idxSizeC,
    int sizeC,
    int free_queue_idxSizeD,
    int sizeD,
    int free_queue_idxOverflow,
    int deviceIdx,
    int processIdx,
    int error_handling
)
{
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idxSizeA & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idxSizeA >> 12);
    lvRxFlowCfg.rx_fdq0_sz1_qnum = (free_queue_idxSizeB & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz1_qmgr = (free_queue_idxSizeB >> 12);
    lvRxFlowCfg.rx_fdq0_sz2_qnum = (free_queue_idxSizeC & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz2_qmgr = (free_queue_idxSizeC >> 12);
    lvRxFlowCfg.rx_fdq0_sz3_qnum = (free_queue_idxSizeD & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz3_qmgr = (free_queue_idxSizeD >> 12);
    /*
     * Other fields are left to be modified by user 
     */
    lvRxFlowCfg.rx_size_thresh0 = 16 * 1024;

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;

    /*
     * Store the handle for future usage
     */
    flowHandleArray[lvArrayIdx][flow_idx] = lvRxFlowHnd;

    {
        CSL_SrioHandle lvHSrio;
        Qmss_Queue queueInfo;
        SRIO_TYPE9_MESSAGE message9;

        lvHSrio = CSL_SRIO_Open (0);

        message9.cos = 0;
        message9.cosMask = 0xFF;
        message9.tt = 1;
        message9.srcProm = 0x1;
        message9.srcId = 0x0;
        message9.dstProm = 0x1;
        message9.dstId = deviceIdx;
        message9.streamId = processIdx << 14;
        message9.streamMask = 0xC000;   /* Only the 2 MSB of the streamIs are compared  */
        message9.flowId = (uint8_t) Cppi_getFlowId (lvRxFlowHnd);

        /*
         * Get the Queue Information for the Receive Completion Queue 
         */
        queueInfo = Qmss_getQueueNumber (dst_queue_idx);

        /*
         * Create the mapping 
         */
        CSL_SRIO_MapType9MessageToQueue (lvHSrio, 0, &message9, queueInfo.qNum);
    }
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_chain_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idxSizeA,
    int sizeA,
    int free_queue_idxSizeB,
    int sizeB,
    int free_queue_idxSizeC,
    int sizeC,
    int free_queue_idxSizeD,
    int sizeD,
    int free_queue_idxOverflow,
    int error_handling
)
{
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_dest_tag_lo_sel = 4;
    lvRxFlowCfg.rx_dest_tag_hi_sel = 5;
    lvRxFlowCfg.rx_src_tag_hi_sel = 4;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idxSizeA & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idxSizeA >> 12);
    lvRxFlowCfg.rx_fdq0_sz1_qnum = (free_queue_idxSizeB & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz1_qmgr = (free_queue_idxSizeB >> 12);
    lvRxFlowCfg.rx_fdq0_sz2_qnum = (free_queue_idxSizeC & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz2_qmgr = (free_queue_idxSizeC >> 12);
    lvRxFlowCfg.rx_fdq0_sz3_qnum = (free_queue_idxSizeD & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz3_qmgr = (free_queue_idxSizeD >> 12);
    lvRxFlowCfg.rx_fdq1_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq1_qmgr = (free_queue_idxOverflow >> 12);
    lvRxFlowCfg.rx_fdq2_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq2_qmgr = (free_queue_idxOverflow >> 12);
    lvRxFlowCfg.rx_fdq3_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq3_qmgr = (free_queue_idxOverflow >> 12);
 
    /* Check if thresholds need to be used and set them accordingly to queue sizes */
    if (free_queue_idxSizeB == free_queue_idxSizeA)
	goto skip_threshold;

    lvRxFlowCfg.rx_size_thresh0_en = 1;
    lvRxFlowCfg.rx_size_thresh0    = sizeA;
    
    if (free_queue_idxSizeC == free_queue_idxSizeB)
	goto skip_threshold;
    
    lvRxFlowCfg.rx_size_thresh1_en = 1;
    lvRxFlowCfg.rx_size_thresh1    = sizeB;
    
    if (free_queue_idxSizeD == free_queue_idxSizeC)
	goto skip_threshold;
    
    lvRxFlowCfg.rx_size_thresh2_en = 1;
    lvRxFlowCfg.rx_size_thresh2    = sizeC;

skip_threshold:

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    /*
     * Store the handle for future usage
     */
    flowHandleArray[lvArrayIdx][flow_idx] = lvRxFlowHnd;

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_tx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_TxChInitCfg lvTxChCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_ChHnd lvTxChHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvTxChCfg, 0, sizeof (Cppi_TxChInitCfg));

    lvTxChCfg.channelNum = channel_idx;
    lvTxChCfg.priority = 2;
    lvTxChCfg.txEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Tx Channel 
     */
    lvTxChHnd = (Cppi_ChHnd) Cppi_txChannelOpen ((Cppi_Handle) dma_idx, &lvTxChCfg, &lvIsAllocated);

    txChannelHandleArray[lvArrayIdx][channel_idx] = lvTxChHnd;

    // The TX Channel shall not be already used by another application.
    if ((lvTxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_tx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx;
    Cppi_ChHnd lvTxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvTxChHnd = txChannelHandleArray[lvArrayIdx][lvChannelIdx];
    // Current XGE implementation workaround
    if (lvTxChHnd == (Cppi_ChHnd) 0)
    {
        lvChannelIdx += XGE_FLOW_IDX_START;
        lvTxChHnd = txChannelHandleArray[lvArrayIdx][lvChannelIdx];
    }

    lvRes = Cppi_channelClose (lvTxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    txChannelHandleArray[lvArrayIdx][lvChannelIdx] = (Cppi_ChHnd) 0;
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_tx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_TxChInitCfg lvTxChCfg;
    Cppi_ChHnd lvTxChHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvTxChCfg, 0, sizeof (Cppi_TxChInitCfg));

    lvTxChCfg.channelNum = XGE_FLOW_IDX_START + channel_idx;
    lvTxChCfg.priority = 2;
    lvTxChCfg.txEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Tx Channel 
     */
    lvTxChHnd = (Cppi_ChHnd) Cppi_txChannelOpen ((Cppi_Handle) dma_idx, &lvTxChCfg, &lvIsAllocated);

    txChannelHandleArray[lvArrayIdx][lvTxChCfg.channelNum] = lvTxChHnd;

    // The TX Channel shall not be already used by another application.
    if ((lvTxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_tx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvTxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvTxChHnd = txChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelClose (lvTxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    txChannelHandleArray[lvArrayIdx][lvChannelIdx] = (Cppi_ChHnd) 0;
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */
em_status_t
ti_em_rio_tx_channel_open (
    int dma_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_TxChInitCfg txCfg;
    Cppi_ChHnd txChHnd;
    uint8_t isAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;


    /*
     * Open the SRIO Transmit Channel and keep it disabled. 
     */
    txCfg.channelNum = CPPI_PARAM_NOT_SPECIFIED;
    txCfg.priority = Srio_PktDma_Prio_Low;
    txCfg.txEnable = Cppi_ChState_CHANNEL_ENABLE;
    txCfg.filterEPIB = 0;
    txCfg.filterPS = 0;
    txCfg.aifMonoMode = 0;
    txChHnd = Cppi_txChannelOpen ((Cppi_Handle) dma_idx, &txCfg, &isAllocated);
    if (txChHnd == NULL)
    {
        return (em_status_t) EM_ERR;
    }
    txChannelHandleArray[lvArrayIdx][0] = txChHnd;
    return (em_status_t) EM_OK;
}


/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_RxChInitCfg lvRxChCfg;
    Cppi_ChHnd lvRxChHnd;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvRxChCfg, 0, sizeof (Cppi_RxChInitCfg));

    lvRxChCfg.channelNum = channel_idx;
    lvRxChCfg.rxEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Rx Channel 
     */
    lvRxChHnd = Cppi_rxChannelOpen ((Cppi_Handle) dma_idx, &lvRxChCfg, &lvIsAllocated);

    rxChannelHandleArray[lvArrayIdx][channel_idx] = lvRxChHnd;

    // The RX Channel shall not be already used by another application.
    if ((lvRxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx;
    Cppi_ChHnd lvRxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxChHnd = rxChannelHandleArray[lvArrayIdx][lvChannelIdx];
    // XGE current implementation workaround
    if (lvRxChHnd == (Cppi_ChHnd) 0)
    {
        lvChannelIdx += XGE_FLOW_IDX_START;
        lvRxChHnd = rxChannelHandleArray[lvArrayIdx][lvChannelIdx];
    }

    lvRes = Cppi_channelClose (lvRxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    rxChannelHandleArray[lvArrayIdx][lvChannelIdx] = (Cppi_ChHnd) 0;
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_RxChInitCfg lvRxChCfg;
    Cppi_ChHnd lvRxChHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvRxChCfg, 0, sizeof (Cppi_RxChInitCfg));

    lvRxChCfg.channelNum = XGE_FLOW_IDX_START + channel_idx;
    lvRxChCfg.rxEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Rx Channel 
     */
    lvRxChHnd = Cppi_rxChannelOpen ((Cppi_Handle) dma_idx, &lvRxChCfg, &lvIsAllocated);

    // The RX Channel shall not be already used by another application.
    if ((lvRxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;

    rxChannelHandleArray[lvArrayIdx][lvRxChCfg.channelNum] = lvRxChHnd;

    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvRxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxChHnd = rxChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelClose (lvRxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    rxChannelHandleArray[lvArrayIdx][lvChannelIdx] = (Cppi_ChHnd) 0;
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */
ti_em_queue_id_t
ti_em_xge_tx_queue_base_idx_get (
)
{
    return (ti_em_queue_id_t) (800 + XGE_FLOW_IDX_START);
}



/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_xge_tx_queue_open (
    ti_em_queue_id_t queue_base_idx,
    int    vlan_priority
)
{
    em_status_t lvRetStatus;
    /*
     * open xge Tx DMA Tx queue 
     */
    lvRetStatus = ti_em_hw_queue_open (queue_base_idx + vlan_priority);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;
    return (ti_em_queue_id_t) (queue_base_idx + vlan_priority);
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_get_absolute_queue_id (
    ti_em_dma_id_t dma_idx,
    ti_em_queue_id_t queue_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;

    if (dmaObjPtr->dmaNum == Cppi_CpDma_QMSS_CPDMA)
        return (ti_em_queue_id_t) (queue_idx + QMSS_INFRASTRUCTURE_QUEUE_BASE);
    else
        return (ti_em_queue_id_t) (queue_idx + QMSS_INFRASTRUCTURE_QUEUE_BASE + 8192);
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_rio_tx_queue_open (
    int rio_tx_queue_idx
)
{
    em_status_t lvRetStatus;
    /*
     * open SRIO DMA Tx queue 
     */
    lvRetStatus = ti_em_hw_queue_open (rio_tx_queue_idx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;
    return (ti_em_queue_id_t) (rio_tx_queue_idx);
}

em_status_t
ti_em_xge_rx_channel_disable (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvRxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxChHnd = rxChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelDisable (lvRxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    return (em_status_t) EM_OK;
}

em_status_t
ti_em_xge_tx_channel_disable (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvTxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvTxChHnd = txChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelDisable (lvTxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    return (em_status_t) EM_OK;
}
em_status_t
ti_em_xge_rx_channel_enable (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvRxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxChHnd = rxChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelEnable (lvRxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    return (em_status_t) EM_OK;
}

em_status_t
ti_em_xge_tx_channel_enable (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    int lvChannelIdx = channel_idx + XGE_FLOW_IDX_START;
    Cppi_ChHnd lvTxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvTxChHnd = txChannelHandleArray[lvArrayIdx][lvChannelIdx];

    lvRes = Cppi_channelEnable (lvTxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;

    return (em_status_t) EM_OK;
}


